// SPDX-License-Identifier: GPL-3.0-or-later
// Copyright © 2022 Florian Fischer
#include <err.h>
#include <fcntl.h>
#include <liburing.h>
#include <pthread.h>

#include <cassert>
#include <chrono>
#include <filesystem>
#include <iostream>
#include <string>
#include <thread>
#include <vector>

static const unsigned URING_ENTRIES = 8;
static const unsigned URING_FILE_TABLE_SIZE = 50;

namespace fs = std::filesystem;
std::vector<std::string*> paths;

const unsigned nthreads = std::thread::hardware_concurrency();

static void collect_paths(const std::string& root) {
	for (const auto& p : fs::recursive_directory_iterator(root))
		paths.emplace_back(new std::string(p.path()));
}

static auto submit_and_wait(struct io_uring* ring) -> int {
#ifdef NDEBUG
	__attribute__((unused))
#endif
	int submitted = io_uring_submit_and_wait(ring, 1);
	assert(submitted == 1);

	struct io_uring_cqe* cqe;
	if (io_uring_peek_cqe(ring, &cqe) < 0) err(EXIT_FAILURE, "failed to peek cqe");

	int res = cqe->res;
	if (res < 0) {
		errno = -res;
		err(EXIT_FAILURE, "opening failed");
	}

	io_uring_cqe_seen(ring, cqe);
	return res;
}

static void close_fd(struct io_uring* ring, int fd) {
	struct io_uring_sqe* sqe = io_uring_get_sqe(ring);
	assert(sqe);

	io_uring_prep_close(sqe, fd);
	sqe->flags |= IOSQE_CQE_SKIP_SUCCESS;

#ifdef NDEBUG
	__attribute__((unused))
#endif
	int submitted = io_uring_submit(ring);
	assert(submitted == 1);
}

static void open_global(struct io_uring* ring, const char* path) {
	struct io_uring_sqe* sqe = io_uring_get_sqe(ring);
	assert(sqe);
	io_uring_prep_openat(sqe, AT_FDCWD, path, O_RDONLY, 0);

	int fd = submit_and_wait(ring);
	close_fd(ring, fd);
}

static void open_direct(struct io_uring* ring, const char* path, unsigned slot) {
	struct io_uring_sqe* sqe = io_uring_get_sqe(ring);
	assert(sqe);
	io_uring_prep_openat_direct(sqe, AT_FDCWD, path, O_RDONLY, 0, slot);

	submit_and_wait(ring);
}

static void thread_func(pthread_barrier_t& init_barrier, unsigned id, bool direct) {
	struct io_uring ring;
	if (io_uring_queue_init(URING_ENTRIES, &ring, 0) < 0)
		err(EXIT_FAILURE, "%d: io_uring init failed", id);

	if (direct) {
		std::vector<int> files(URING_FILE_TABLE_SIZE, -1);
		if (io_uring_register_files(&ring, files.data(), files.size()) < 0)
			err(EXIT_FAILURE, "%d: creating empty file table failed", id);
	}

	const auto total_paths = paths.size();
	auto paths_to_open = total_paths / nthreads;
	const auto paths_offset = paths_to_open * id;
	if (id == nthreads - 1) paths_to_open += total_paths % paths_to_open;

	pthread_barrier_wait(&init_barrier);

	for (unsigned i = 0; i < paths_to_open; ++i) {
		const auto* p = paths[paths_offset + i];
		assert(p->c_str());

		if (direct) {
			open_direct(&ring, p->c_str(), i % URING_FILE_TABLE_SIZE);
		} else {
			open_global(&ring, p->c_str());
		}
	}
}

static void bench_open(bool direct) {
	pthread_barrier_t open_latch;
	pthread_barrier_init(&open_latch, nullptr, nthreads + 1);
	std::vector<std::thread> threads;
	for (unsigned i = 0; i < nthreads; ++i)
		threads.emplace_back([&, i = i] { thread_func(open_latch, i, direct); });

	pthread_barrier_wait(&open_latch);

	auto start = std::chrono::high_resolution_clock::now();
	for (auto& t : threads) t.join();
	auto stop = std::chrono::high_resolution_clock::now();
	auto duration = std::chrono::duration_cast<std::chrono::nanoseconds>(stop - start);

	std::cout << (direct ? "directly " : "") << "opening " << paths.size()
						<< " files took: " << duration.count() << "ns" << std::endl;
}

auto main(int argc, char* argv[]) -> int {
	std::string root = ".";
	bool direct = false;
	if (argc > 1) {
		direct = std::string(argv[1]) == "--direct";
		root = std::string(argv[argc - 1]);
	}

	collect_paths(root);

	bench_open(direct);
	return 0;
}
